# import re
import os
import json

resources = {}
# Explanation: http://wiki.dbpedia.org/Downloads2015-04
# Link to this set: http://downloads.dbpedia.org/2015-04/core-i18n/en/instance-types_en.nt.bz2
with open('/Users/pascal/Downloads/instance_types_en.nt') as f: 
    for line in f:
        try:
            parts = line.split()
            first_part = parts[0]
            if first_part.count('/') == 4:
                name = first_part.split('/')[-1].replace('_', ' ').lower()[:-1]
                if len(name.split()) > 3:
                    continue
                pos = name.find('  ')
                if pos > -1:
                    name = name[:pos]       
                if 'ontologydesignpatterns' in parts[2]:                    
                    continue
                if 'w3.org' in parts[2]:                    
                    continue
                if name not in resources: 
                    resources[name] = set() 
                resources[name].add(parts[2].split('/')[-1][:-1]) 
        except IndexError:
            pass    

# resources start with 'http://dbpedia.org/resource/'

# types = set()
# for name in resources:
#     m = re.search(r'(\(.+\))', name)
#     if m:
#         types.add(m.groups()[0].lower())            

stripped_resources = {}
for x in resources:
    if any([y in x for y in '0123456789']): 
        if '%' not in x: 
            continue
    pos = x.find('(') 
    if pos > -1:
        stripped_resources[x[:pos].strip()] = list(resources[x])
    else:    
        stripped_resources[x.strip()] = list(resources[x])

#######################################################################################################################        

fname = os.path.join(os.path.dirname(__file__), 'dbpedia.json')
with open(fname, 'w') as f: 
    json.dump(stripped_resources, f)

# from ZODB.FileStorage import FileStorage
# from ZODB.DB import DB

# import transaction
# from BTrees.OOBTree import OOBTree
        
# storage = FileStorage('/Users/pascal/GDrive/appear/Data.fs')
# db = DB(storage)
# connection = db.open()
# root = connection.root()

# root['dbpedia'] = OOBTree(stripped_resources)
# transaction.commit()



# Link('http://nu.nl/nieuws', slugified = 'httpnunlnieuws', parent = 'http://nu.nl', hop = 1, refTo, refFrom)

import re

def get_word_set(url): 
    question_mark_index = url.find('?')
    url = url[:question_mark_index] if question_mark_index > -1 else url
    return set(re.split('[^a-zA-Z0-9]', url))

def get_similarity(url1, ws2, sc2):
    ws1 = get_word_set(url1) 
    if ws1 == ws2:
        return (0, 0)
    return - abs(url1.count('/') - sc2), len(ws1 & ws2) / len(ws2)
    
def get_sorted_links(urls, request_url):
    wsr = get_word_set(request_url)
    slash_count = request_url.count('/')
    return sorted(set(urls), key = lambda u: get_similarity(u, wsr, slash_count), reverse = True)



from sky.configs import DEFAULT_CRAWL_CONFIG
from sky.scraper import Scraper

# Crawling
CRAWL_CONFIG = DEFAULT_CRAWL_CONFIG
CRAWL_CONFIG.update({
    'seed_urls': [
        'http://www.techcrunch.com/2015/05/21/chromebook-sales-predicted-to-grow-by-27-this-year-to-7-3m-units/'
    ],

    'collections_path': '/Users/pascal/GDrive/sky_collections',

    'collection_name': 'techie',

    # Optional

    'crawl_filter_regexps': [

    ],

    'crawl_required_regexps': [
        '2015', '2014'
    ],

    'index_filter_regexps': [

    ],

    'index_required_regexps': [
        '2015', '2014'
    ],

    'max_saved_responses': 100,

    'max_workers': 10,
})

from sky.crawler import crawl
crawl.start(CRAWL_CONFIG)

# Indexing

SCRAPE_CONFIG = CRAWL_CONFIG.copy()

SCRAPE_CONFIG.update({
    'template_proportion': 0.09,
    'max_templates': 1000
})

skindex = Scraper(SCRAPE_CONFIG)

skindex.load_local_pages()
skindex.add_template_elements()

res = skindex.process_all(remove_visuals=True)



# For the body:
# - Use and strip HTML 5 elements
# - Cut-off text when we see a clear "share elements" thing

# Really use templates/content things, for example to detect "sharing",
# "comments", "relevant links"

# logging all actions
# simplify

from __future__ import absolute_import

import re
import os
import json
import lxml.html
import justext

# from sky.dbpedia import load_dbpedia
# from sky.dbpedia import get_dbpedia_from_words
from sky.money import MoneyMatcher

# dbpedia = load_dbpedia()
money = MoneyMatcher()

from sky.helper import get_text_and_tail
from sky.helper import fscore
from sky.helper import makeTree
from sky.helper import extractDomain
from sky.helper import get_last_text_non_a_node
from sky.helper import normalize


from sky.remove_boilerplate import DomainNodesDict
from sky.findTitle import getRuleTitle
from sky.get_date import get_dates
from sky.get_author import get_author
from sky.links import get_sorted_links
from sky.language import get_language
from sky.images import get_images


class Scraper:
    # todo is finextr date

    def __init__(self, config):
        self.config = config
        self.index_required_regexps = []
        self.index_filter_regexps = []
        self.detected_language = None
        self.seed_urls = None
        self.domain = None
        self.collections_path = None
        self.collection_name = None
        self.template_proportion = None
        self.min_templates = None
        self.max_templates = None
        self.applyConfigFile()
        self.domain = extractDomain(self.seed_urls[0])
        self.url_to_tree_mapping = {}
        self.url_to_headers_mapping = {}

        # Boilerplate remover class
        self.domain_nodes_dict = DomainNodesDict(
            self.domain, self.min_templates, self.max_templates, self.template_proportion)
        if 'template_dict' in config:
            self.domain_nodes_dict.update(dict(config['template_dict']))
            vals = self.domain_nodes_dict.values()
            self.domain_nodes_dict.num_urls = max(vals) if vals else 0

    def remove_bad_xpath_from_tree(self, tree, bad_xpath):
        bad_element = tree.find('.' + bad_xpath)
        while bad_element is not None:
            bad_element.getparent().remove(bad_element)
            bad_element = tree.find('.' + bad_xpath)

    def should_save(self, url):
        if (not self.index_required_regexps or
                any([re.search(condition, url) for condition in self.index_required_regexps])):
            if all([not re.search(x, url) for x in self.index_filter_regexps]):
                return True
        return False

    def applyConfigFile(self):
        for config_key, config_value in self.config.items():
            setattr(self, config_key, config_value)
        self.domain = extractDomain(self.seed_urls[0])

    # This might have to be changed into the database variant
    def load_local_pages(self):
        saved_html_dir = os.path.join(self.collections_path, self.collection_name)
        for _, _, files in os.walk(saved_html_dir):
            for name in files:
                if name.startswith('.DS_'):
                    continue
                with open(os.path.join(saved_html_dir, name)) as f:
                    try:
                        js = json.load(f)
                    except:
                        print('failed to load json {}'.format(name))
                        continue
                    try:
                        self.url_to_tree_mapping[js['url']] = makeTree(
                            js['html'], self.domain)
                        self.url_to_headers_mapping[js['url']] = js['headers']
                    # pylint: disable=broad-except
                    except Exception as e:
                        print(str(e))

    def add_template_elements(self):
        for url in self.url_to_tree_mapping:
            self.domain_nodes_dict.add_template_elements(self.url_to_tree_mapping[url])

    def remove_bad_xpaths_from_tree(self, tree):
        if 'bad_xpaths' in self.config:
            for bad_xpath in self.config['bad_xpaths']:
                self.remove_bad_xpath_from_tree(tree, bad_xpath)

    def process(self, url, tree, remove_visuals, exclude_data):
        self.remove_bad_xpaths_from_tree(tree)

        if self.detected_language is None:
            self.detected_language = get_language(
                tree, self.url_to_headers_mapping[url], self.domain)
        # print('language: {}'.format(self.detected_language))
        # pre_text_content = normalize('\n'.join([get_text_and_tail(x) for x in tree.iter()]))

        # author has to be attempted before duplicate removal, since an author is
        # likely to occur more often
        self.domain_nodes_dict.remove_template(tree)
        hardest_authors, not_hardest_authors, text_hard_authors, text_soft_authors, meta_authors = get_author(
            tree, self.detected_language)
        self.domain_nodes_dict.remove_author(tree)
        title = getRuleTitle(tree)
        # filter duplicate images by src
        ok_imgs = get_images(tree)
        titleind = ()
        imginds = []
        contentinds = []

        # such as title, date and later author

        link_eles = [link[0] for link in tree.iterlinks()
                     if link[0].tag == 'a' and link[2] and
                     link[2].startswith(self.domain) and
                     get_text_and_tail(link[0]).strip()]

        linkinds = []
        for num, node in enumerate(tree.iter()):
            if node in ok_imgs:
                imginds.append((node, num))
            elif normalize(get_text_and_tail(node)) == title:
                titleind = (node, num)
            elif get_text_and_tail(node).strip():
                if node in link_eles:
                    linkinds.append((node, num))
                contentinds.append((node, num))
            # Cleanup trash for visual'
            if remove_visuals:
                if node.tag == 'input':
                    node.set('type', 'hidden')
                elif node.tag == 'a' and not get_text_and_tail(node).strip():
                    for att in node.attrib:
                        node.set(att, '')
                if node.tag == 'img':
                    node.set('alt', '')
                if node.attrib and 'background-image' in node.attrib:
                    node.set('background-image', '')
        if not titleind:
            # fuzzy token text / title matching
            title_set = set(title.split())
            for num, node in enumerate(tree.iter()):
                text_content = get_text_and_tail(node)
                if text_content and len(text_content) < 500:
                    text_set = set(text_content.split())
                    if fscore(title_set, text_set) > 0.5:
                        titleind = (node, num)
                        break

        if titleind:
            sortedimgs = sorted(imginds, key=lambda x: abs(x[1] - titleind[1]))
        else:
            sortedimgs = []
        hardest_dates, fuzzy_hardest_dates, not_hardest_dates, soft_dates, non_fuzzy_anys, fuzzy_anys = get_dates(
            tree, self.detected_language)
        images = []
        for x in sortedimgs:
            val = None
            if 'src' in x[0].attrib:
                val = x[0].attrib['src']
            elif 'content' in x[0].attrib:
                val = x[0].attrib['content']
            elif 'style' in x[0].attrib:
                tmp = re.findall(r'background-image:[ ]*url\((http[^)]+)', x[0].attrib['style'])
                if tmp:
                    val = tmp[0]
            if val is not None and val not in images:
                images.append(val)

        date = ''
        date_node_index = None
        author = ''
        author_node_index = None
        if titleind:
            # excluding soft dates (meta, they wont work anyway)
            for dt in [hardest_dates, fuzzy_hardest_dates, not_hardest_dates,
                       non_fuzzy_anys, fuzzy_anys]:
                if dt:
                    date, date_node_index = sorted(dt, key=lambda x: abs(x[1] - titleind[1]))[0]
                    break
            for at in [hardest_authors, not_hardest_authors, text_hard_authors, text_soft_authors]:
                if at:
                    author, author_node_index = sorted(
                        at, key=lambda x: abs(x[1] - titleind[1]))[0]
                    break
        if not date and soft_dates:
            for sd in soft_dates:
                date = sd
                break

        if not date:
            date = "1970-01-01"

        if not author and meta_authors:
            for ma in meta_authors:
                author = ma
                break

        if date_node_index is not None:
            for num, node in enumerate(tree.iter()):
                if num == date_node_index:
                    break
            # It goes wrong when some year is mentioned in the title, then it removes title
            # print('removing date content', node.text)
            node.text = ''
            node.tail = ''

        if author_node_index is not None:
            for num, node in enumerate(tree.iter()):
                if num == author_node_index:
                    break
            # It goes wrong when some year is mentioned in the title, then it removes title
            # print('removing author content', node.text)
            node.text = ''
            node.tail = ''

        cleaned_html = lxml.html.tostring(tree).decode('utf8')

        body_content = self.get_content(cleaned_html)

        if not body_content:
            body_content = []
            title_len = len(title)
            title_tokens = set(title.split())
            len_title_tokens = len(title_tokens)
            last_text_node_num = get_last_text_non_a_node(tree)
            for num, x in enumerate(tree.iter()):
                txt = normalize(get_text_and_tail(x))
                if txt:
                    if num < titleind[1]:
                        # print('removed pre-title', txt)
                        x.text = ''
                        x.tail = ''
                        continue
                    if last_text_node_num > 0 and num > last_text_node_num:
                        # print('removed post-content', txt)
                        x.text = ''
                        continue
                    n = len(txt)
                    # remove title
                    txt_tokens = set(txt.split())
                    n_matching = len(txt_tokens & title_tokens)
                    if (n < title_len * 3 and n_matching / len(txt_tokens) > 0.3 and
                            n_matching / len_title_tokens > 0.3):
                        # print('removed!', txt)
                        continue
                    body_content.append(txt)

        links = [x.attrib['href'] for x in tree.xpath('//a')
                 if 'href' in x.attrib and
                 x.attrib['href'].startswith(self.domain) and
                 self.should_save(x.attrib['href'])]

        money_amounts = money.find('\n'.join(body_content), 1000) + money.find(title, 1000)

        data = {'title': title,
                'body': body_content,
                'images': images,
                'publish_date': str(date),
                'author': author,
                'cleaned': cleaned_html,
                'language': self.detected_language,
                'url': url,
                'domain': self.domain,
                'money': money_amounts,
                'summary': '',
                'related': get_sorted_links(links, url)[:5]}

        filtered_data = {k: v for k, v in data.items() if k not in exclude_data}

        return filtered_data

    def process_all(self, remove_visuals=False, exclude_data=None, maxn=100000000):
        if exclude_data is None:
            exclude_data = []
        results = {}
        for num, url in enumerate(self.url_to_tree_mapping):
            if num > maxn:
                break
            results[url] = self.process(url, self.url_to_tree_mapping[url],
                                        remove_visuals, exclude_data)
        return results

    def get_content(self, html):
        # I should refactor the other get_content when this fails into here
        lang_mapping = {'nl': 'Dutch', 'en': 'English', 'com': 'English'}
        if self.detected_language not in lang_mapping:
            return ''
        lang = lang_mapping[self.detected_language]
        body_content = [x.text for x in justext.justext(html, justext.get_stoplist(lang))
                        if not x.is_boilerplate and not x.is_heading]
        return body_content



# -*- coding: utf-8 -*-

import tornado.web
import tornado.autoreload
import tornado

import os
import shutil

from sky.crawler import crawl
from sky.configs import DEFAULT_CRAWL_CONFIG
from sky.helper import extractDomain
from sky.scraper import Scraper

# from textblob import TextBlob


def is_numeric(x):
    try:
        int(x)
        return True
    except ValueError:
        return False


class MainHandler(tornado.web.RequestHandler):

    def get(self):
        self.render('page_template.html', items=[], cached=False)

    def post(self):
        CRAWL_CONFIG = DEFAULT_CRAWL_CONFIG
        CRAWL_CONFIG.update({
            'collections_path': os.path.join(os.path.expanduser('~'), 'sky_collections/'),
            # 'max_workers': 10,
        })
        args = self.request.arguments
        print(args)
        for arg in args:
            value = args[arg][0].decode('utf8')
            if value and arg != 'url' and arg != 'checkboxcache':
                print('pre', arg, CRAWL_CONFIG[arg])
                if isinstance(CRAWL_CONFIG[arg], list):
                    CRAWL_CONFIG[arg] = [int(value)] if is_numeric(value) else [value]
                else:
                    CRAWL_CONFIG[arg] = int(value) if is_numeric(value) else value
                print('post', arg, CRAWL_CONFIG[arg])

        url = self.get_argument('url', '')

        use_cache = self.get_argument('checkboxcache', '')

        domain = extractDomain(url)
        CRAWL_CONFIG['seed_urls'] = [url]
        CRAWL_CONFIG['collection_name'] = domain[7:]

        if use_cache != 'on':

            col_path = os.path.join(CRAWL_CONFIG['collections_path'],
                                    CRAWL_CONFIG['collection_name'])
            print(col_path)
            if os.path.exists(col_path):
                shutil.rmtree(col_path)

            crawl.start(CRAWL_CONFIG)

        SCRAPE_CONFIG = CRAWL_CONFIG.copy()

        SCRAPE_CONFIG.update({
            'template_proportion': 0.4,
            'max_templates': 100
        })

        skindex = Scraper(SCRAPE_CONFIG)

        skindex.load_local_pages()
        skindex.add_template_elements()

        res = skindex.process_all(remove_visuals=True,
                                  maxn=CRAWL_CONFIG['max_saved_responses'])

        items = []
        for num, url in enumerate(res):
            if num == CRAWL_CONFIG['max_saved_responses']:
                break
            dc = res[url]
            dc['url'] = url
            dc['source_name'] = domain
            dc['images'] = [x for x in reversed(dc['images'][:5])]
            # dc['blobs'] = [TextBlob(x) for x in dc['body'] if dc['body']]
            items.append(dc)

        if items and 'money' in items[0]:
            items = sorted(items, key=lambda x: len(x['money']), reverse=True)

        self.render('page_template.html', items=items, cached=False)

settings = {
    'template_path': os.path.join(os.path.dirname(__file__), 'templates'),
    'static_path': os.path.join(os.path.dirname(__file__), 'static')
}


def main(host='localhost', port=7900):
    # to run the server, type-in $ python view.py

    application = tornado.web.Application([
        (r"/", MainHandler),
    ], **settings)

    application.listen(int(port), host)

    ioloop = tornado.ioloop.IOLoop().instance()

    print('serving skyViewer at "{}:{}" from file: {}'.format(host, port, __file__))

    ioloop.start()

if __name__ == '__main__':
    main()



import re

from sky.helper import normalize
from sky.helper import get_text_and_tail


class DomainNodesDict(dict):

    def __init__(self, domain, min_templates=None, max_templates=None, template_proportion=None):
        super(DomainNodesDict, self).__init__()
        self.num_urls = 0
        self.domain = domain
        self.min_templates = min_templates or 2
        self.max_templates = max_templates or 24
        self.template_proportion = template_proportion or 0.7
        self.untemplated = []

    def get_fingerprints(self, node):
        res = []
        text = normalize(get_text_and_tail(node)).strip()
        if node.tag == 'a' and 'href' in node.attrib:
            res = [(node.tag, node.attrib['href'], '', '')]
        if text:
            res += [(node.tag, a, node.attrib[a], text) for a in node.attrib]
            if node.tag == 'a':
                res += [(node.tag, '', '', text)]
            if not res:
                res = [(node.tag, '', '', text)]
        else:
            if 'style' in node.attrib:
                tmp = re.findall(r'background-image:[ ]*url\((http[^)]+)', node.attrib['style'])
                if tmp:
                    res += [('style', tmp[0])]
        return res

    def add_fp(self, fp, seen):
        if fp not in seen:
            if fp not in self:
                self[fp] = 0
            self[fp] += 1
            seen.add(fp)

    def add_template_elements(self, tree):
        if self.num_urls < self.max_templates:
            seen = set()
            for node in tree.iter():
                if (node.tag == 'meta' and 'property' in node.attrib and
                        'image' in node.attrib['property'] and 'content' in node.attrib):
                    self.add_fp((node.attrib['property'], node.attrib['content']), seen)
                elif node.tag in ['img', 'iframe'] and 'src' in node.attrib:
                    self.add_fp((node.tag, node.attrib['src']), seen)
                else:
                    for fp in self.get_fingerprints(node):
                        self.add_fp(fp, seen)
            self.num_urls += 1

    def possible_author(self, node):
        goods = ['author', 'by', 'publi', 'write', 'written', 'info']
        attr_values = node.attrib.values()
        if any([g in a for a in attr_values for g in goods]):
            return True
        txt = get_text_and_tail(node)
        if re.search(r'\b(author|by)[: ]', txt):
            return True
        return False

    def remove_template(self, tree):
        if self.num_urls < self.min_templates:
            return False
        for node in tree.iter():
            if self.possible_author(node):
                continue
            if node.tag == 'meta':
                if 'property' in node.attrib and 'content' in node.attrib:
                    fp = (node.attrib['property'], node.attrib['content'])
                    if fp in self and self[fp] / self.num_urls > self.template_proportion:
                        node.set('content', '')
            elif node.tag in ['img', 'iframe'] and 'src' in node.attrib:
                fp = (node.tag, node.attrib['src'])
                if fp in self and self[fp] / self.num_urls > self.template_proportion:
                    node.set('src', '')
                    node.set('alt', '')
            else:
                for fp in self.get_fingerprints(node):
                    if fp in self and self[fp] / self.num_urls > self.template_proportion:
                        node.text = ''
                        node.tail = ''
                        if node.tag == 'a':
                            for child in node.iter():
                                child.text = ''
                                child.tail = ''

        return True

    def remove_author(self, tree):
        for node in tree.iter():
            if self.possible_author(node):
                if node.tag == 'meta':
                    if 'property' in node.attrib and 'content' in node.attrib:
                        fp = (node.attrib['property'], node.attrib['content'])
                        if fp in self and self[fp] / self.num_urls > self.template_proportion:
                            node.set('content', '')
                elif node.tag in ['img', 'iframe'] and 'src' in node.attrib:
                    fp = (node.tag, node.attrib['src'])
                    if fp in self and self[fp] / self.num_urls > self.template_proportion:
                        node.set('src', '')
                        node.set('alt', '')
                else:
                    for fp in self.get_fingerprints(node):
                        if fp in self and self[fp] / self.num_urls > self.template_proportion:
                            node.text = ''
                            node.tail = ''
                            if node.tag == 'a':
                                for child in node.iter():
                                    child.text = ''
                                    child.tail = ''



# adformatie.nl
#   File "/Library/Frameworks/Python.framework/Versions/3.4/lib/python3.4/site-packages/sky/images.py", line 57, in get_images
#     if general_ok_img(img_candidate, wrong_atts):
#   File "/Library/Frameworks/Python.framework/Versions/3.4/lib/python3.4/site-packages/sky/images.py", line 20, in general_ok_img
#     if tmp[0]:
# IndexError: list index out of range


# <figure> should maybe use some weighting
# background-url should also be allowed

# Also return most likely logo (logo, favicon??, first image?, icon)
import re


def general_ok_img(img_candidate, wrong_imgs):
    link = None
    if img_candidate.tag == 'img':
        if 'src' not in img_candidate.attrib:
            return False
        link = img_candidate.attrib['src']
    else:
        if 'content' in img_candidate.attrib:
            link = img_candidate.attrib['content']
        elif 'style' in img_candidate.attrib:
            tmp = re.findall(
                r'background-image:[ ]*url\((http[^)]+)', img_candidate.attrib['style'])
            if tmp:
                link = tmp[0]
    if link is None:
        return False
    # if link longer than 1000 chars, drop it
    if len(link) > 1000:
        return False
    # if link does not start with "http", drop it
    if not link.startswith('http'):
        return False
    # if link attributes contain one of the wrong atts, drop it
    if any([any([w in img_candidate.attrib[a] for w in wrong_imgs])
            for a in img_candidate.attrib]):
        return False
    return True


def dimensions_ok(img_candidate):
    try:
        if 'height' in img_candidate.attrib and int(img_candidate.attrib['height']) < 100:
            return False
        if 'width' in img_candidate.attrib and int(img_candidate.attrib['width']) < 100:
            return False
    except ValueError:
        pass
    return True


def get_images(tree, wrong_atts=None):
    if wrong_atts is None:
        wrong_atts = ['adsense', 'icon', 'logo', 'advert', 'toolbar', 'footer', 'layout', 'banner']
        # dit recoden in een tree.iter() loop, en dan ook de "node index"/num noteren hier.
    img_candidates = tree.xpath('//img[string-length(@src) > 3]')
    img_candidates += tree.xpath('//meta[contains(@property, "image")]')
    img_candidates += tree.xpath('//*[contains(@style, "background-image")]')
    leftover = []
    for img_candidate in img_candidates:
        if general_ok_img(img_candidate, wrong_atts):
            if img_candidate.tag == 'img':
                if dimensions_ok(img_candidate):
                    leftover.append(img_candidate)
            else:
                leftover.append(img_candidate)
    return leftover



import lxml.html.diff
import lxml.html
from bs4 import UnicodeDammit
import tldextract
import re
import requests

from lxml.html.clean import Cleaner

cleaner = Cleaner()
cleaner.javascript = True
cleaner.style = True
#cleaner.kill_tags = ['a', 'h1']
#cleaner.remove_tags = ['p']


def slugify(value):
    return re.sub(r'[^\w\s-]', '', re.sub(r'[-\s]+', '-', value)).strip().lower()


def view_url(url):
    view_html(getQuickTree(url))


def view_html(x):
    import time
    import webbrowser
    fname = '/tmp/' + str(hash(x)) + '.html'
    with open(fname, 'w') as f:
        f.write(x)
    webbrowser.open('file://' + fname)
    time.sleep(1)


def view_node(node, attach_head=False, questionContains=None):
    newstr = makeParentLine(node, attach_head, questionContains)
    view_tree(newstr)


def view_tree(node):
    view_html(lxml.html.tostring(node).decode('utf8'))


def view_diff(obj1, obj2, url='', diffMethod=lxml.html.diff.htmldiff):
    if isinstance(obj1, str):
        tree1 = lxml.html.fromstring(obj1)
        tree2 = lxml.html.fromstring(obj2)
        html1 = obj1
        html2 = obj2
    else:
        html1 = lxml.html.tostring(obj1).decode('utf8')
        html2 = lxml.html.tostring(obj2).decode('utf8')
        tree1 = obj1
        tree2 = obj2
    diffHtml = diffMethod(tree1, tree2)
    diffTree = lxml.html.fromstring(diffHtml)
    insCounts = diffTree.xpath('count(//ins)')
    delCounts = diffTree.xpath('count(//del)')
    pureDiff = ''
    for y in [z for z in diffTree.iter() if z.tag in ['ins', 'del']]:
        if y.text is not None:
            color = 'lightgreen' if 'ins' in y.tag else 'red'
            pureDiff += '<div style="background-color:{};">{}</div>'.format(color, y.text)
    print('From t1 to t2, {} insertions and {} deleted'.format(insCounts, delCounts))
    diff = '<head><title>diff</title><base href="' + url
    diff += '" target="_blank"><style>ins{ background-color:lightgreen; } '
    diff += 'del{background-color:red;}</style></head>' + diffHtml
    view_html(diff)
    view_html(html1)
    view_html(html2)
    view_html('<html><body>{}</body></html>'.format(str(pureDiff)))


def makeParentLine(node, attach_head=False, questionContains=None):
    # Add how much text context is given. e.g. 2 would mean 2 parent's text
    # nodes are also displayed
    if questionContains is not None:
        newstr = doesThisElementContain(questionContains, lxml.html.tostring(node))
    else:
        newstr = lxml.html.tostring(node)
    parent = node.getparent()
    while parent is not None:
        if attach_head and parent.tag == 'html':
            newstr = lxml.html.tostring(parent.find(
                './/head'), encoding='utf8').decode('utf8') + newstr
        tag, items = parent.tag, parent.items()
        attrs = " ".join(['{}="{}"'.format(x[0], x[1]) for x in items if len(x) == 2])
        newstr = '<{} {}>{}</{}>'.format(tag, attrs, newstr, tag)
        parent = parent.getparent()
    return newstr


def extractDomain(url):
    tld = ".".join([x for x in tldextract.extract(url) if x])
    protocol = url.split('//', 1)[0]
    if 'file:' == protocol:
        protocol += '///'
    else:
        protocol += '//'
    return protocol + tld


def doesThisElementContain(text='pagination', nodeStr=''):
    templ = '<div style="border:2px solid lightgreen"><div style="background-color:lightgreen">Does this element contain <b>{}</b>?</div>{}</div>'
    return templ.format(text, nodeStr)


def makeTree(html, domain=None):

    ud = UnicodeDammit(html, is_html=True)
    # tree = lxml.html.fromstring(cleaner.clean_html(ud.unicode_markup),
    #                             base_url = extractDomain(url))
    tree = lxml.html.fromstring(ud.unicode_markup)

    if domain is not None:
        tree.make_links_absolute(domain)

    for el in tree.iter():

        # remove comments
        if isinstance(el, lxml.html.HtmlComment):
            el.getparent().remove(el)
            continue

        if el.tag == 'script':
            el.getparent().remove(el)
            continue

    return tree


def getQuickTree(url, domain=None):
    r = requests.get(url, headers={
        'User-Agent': 'Mozilla/5.0 (Windows NT 7.0; Win64; x64; rv:3.0b2pre) Gecko/20110203 Firefox/4.0b12pre'})
    if domain is None:
        domain = extractDomain(url)
    return makeTree(r.text, domain)


def getLocalTree(url, domain=None):
    if domain is None:
        domain = extractDomain(url)
    with open(url) as f:
        html = f.read()
    return makeTree(html, domain)


def normalize(s):
    return re.sub(r'\s+', lambda x: '\n' if '\n' in x.group(0) else ' ', s).strip()


def get_text_and_tail(node):
    text = node.text if node.text else ''
    tail = node.tail if node.tail else ''
    return text + ' ' + tail


def fscore(x, y):
    try:
        z = sum([w in y for w in x]) / len(x)
        z2 = sum([w in x for w in y]) / len(y)
        return (2 * z * z2) / (z + z2)
    except:
        return 0


def get_pagination(tree):
    links = [(x, x.attrib['href']) for x in tree.xpath('//a') if 'href' in x.attrib]
    res = [re.findall('[0-9]+', x[1]) for x in links]
    for num, (link, x, y, z) in enumerate(zip(links[:-2], res[:-2], res[1:-1], res[2:])):
        if x and y and z:
            if len(x) == len(y) == len(z):
                for i, j, k in zip(x, y, z):
                    if int(i) + 1 == int(j) and int(i) < int(k) and int(j) < int(k):
                        return (link[0], find_common_ancestor(links[num][0], links[num + 1][0]), link[1].replace(i, '{}'))


def find_common_ancestor(n1, n2):
    if n1 is n2:
        return n1.getparent()
    n1_ancestors = set(n1.iterancestors())
    for parent in n2.iterancestors():
        if parent in n1_ancestors:
            return parent


def urlmatcher(url1, url2):
    # can be upgraded to match last part
    tokens1 = url1.split('/')
    tokens2 = url2.split('/')
    l1 = len(tokens1)
    l2 = len(tokens2)
    maxl = max(l1, l2)
    return l1 == l2, sum([t1 == t2 for t1, t2 in zip(tokens1, tokens2)]) / maxl


def get_sorted_similar_urls(tree, url):
    return sorted(tree.xpath('//a/@href'), key=lambda x: (url != x, urlmatcher(url, x)), reverse=True)


def get_last_text_non_a_node(tree):
    for node in reversed(list(tree.iter())):
        if node.tag == 'a':
            continue
        txt = get_text_and_tail(node)
        if txt.strip():
            for p in node.iterancestors():
                if p.tag == 'a':
                    break
            else:
                for num, n in enumerate(tree.iter()):
                    if n == node:
                        return num
    return False



########## 1. Setup ##################################################
PROJECT_NAME = 'testproj'

######### 2a. File      ##############################################

from sky.crawler_services import CrawlFileService
from sky.crawler_plugins import CrawlFilePluginNews

storage_object = {'path': '/Users/pascal/sky_collections/'}

cs = CrawlFileService(PROJECT_NAME, storage_object, CrawlFilePluginNews)

######### 2b. Cloudant  ##############################################

import cloudant
from sky.crawler_services import CrawlCloudantService
from sky.crawler_plugins import CrawlCloudantPluginNews

with open('cloudant.username') as f:
    USERNAME = f.read()
with open('cloudant.password') as f:
    PASSWORD = f.read()
account = cloudant.Account(USERNAME)
account.login(USERNAME, PASSWORD)

cs = CrawlCloudantService(PROJECT_NAME, account, CrawlCloudantPluginNews)

######### 2c. ElasticSearch ##########################################

import elasticsearch

from sky.crawler_services import CrawlElasticSearchService
from sky.crawler_plugins import CrawlElasticSearchPluginNews

es = elasticsearch.Elasticsearch([{'host': 'localhost', 'port': 9200}])

cs = CrawlElasticSearchService(PROJECT_NAME, es, CrawlElasticSearchPluginNews)

######### 2d. ZODB ###################################################

import ZODB.FileStorage
from sky.crawler_services import CrawlZODBService
from sky.crawler_plugins import CrawlZODBPluginNews

fname = '/Users/pascal/sky_collections/zodb/{}.fs'.format(PROJECT_NAME)
storage = ZODB.FileStorage.FileStorage(fname)

cs = CrawlZODBService(PROJECT_NAME, storage, CrawlZODBPluginNews)

######### 3. Add config files to the database ########################
from sky.configs import PRODUCTION_CRAWL_CONFIG

default = cs.get_crawl_plugin('default')
default.save_config(PRODUCTION_CRAWL_CONFIG)

bbc_config = {
    'seed_urls': ['http://www.bbc.com/news/world/europe'],
    'crawl_required_regexps': ['europe'],
    'index_required_regexps': ['news/world-europe-'],
}

bbc = cs.get_crawl_plugin('bbc.com')
bbc.save_config(bbc_config)

######## 4. Start crawling ###########################################
cs['bbc.com']

cs.run('bbc.com')

techcrunch_config = {
    "seed_urls": ["http://www.techcrunch.com"],
    "crawl_required_regexps": [
        "http://techcrunch.com/",
        "http://www.techcrunch.com/"],
    "crawl_filter_regexps": [
        '/video/',
        '/event',
        '.xml',
        'rssfeeds',
        '/contact/'],
    "index_required_regexps": [r"\d{4}/\d{2}/\d{2}/[^/]+/?$"],
    "max_workers": 100,
    "max_saved_responses": 250000
}


techcrunch = cs['techcrunch.com']
techcrunch.save_config(techcrunch_config)
techcrunch.run()



import re

robots_txt = '...'
user_agent_found = False
crawl_filter_regexps = []
crawl_required_regexps = []
for line in robots_txt.split():
    line.strip()
    if line.startswith('#'):
        continue
    if line.startswith('User-agent:'):
        if '*' in line:
            user_agent_found = True
        else:
            user_agent_found = False
    elif user_agent_found:
        if line.startswith('Disallow'):
            res = line.split('Disallow:')
            if res:
                res = res[1:]
                if res == '*':
                    raise ValueError("robots.txt disallows bot to crawl anything.")
            crawl_filter_str = re.escape(line.split('Disallow: ', ))



import os

try:
    from nltk.corpus import stopwords
    stopset = set(stopwords.words('english'))    
except ImportError: 
    print("Cannot use dbpedia without 'pip install nltk'")
    
try:
    import ujson as json
except ImportError:
    import json

def generate_testables(words, stopword_set, n_grams = 4):
    grams = set()
    n = len(words)
    for i in range(len(words)):
        for j in range(n_grams):
            if n - j > i:
                ws = words[i : i + j + 1]
                if any(['NN' not in x[1] for x in ws]):
                    continue
                word_list = [x[0].lower() for x in ws] 
                if any([len(x) < 3 for x in word_list]):
                    continue
                if set(word_list) & stopword_set:
                    continue 
                
                grams.add((" ".join([x[1] for x in ws]), " ".join(word_list)))
    return grams            

def get_dbpedia_from_words(pos_tags, db_dict, ok_entities = None):
    if ok_entities is None:
        ok_entities = ['Person', 'Organisation']
    ws = generate_testables(pos_tags, stopset)
    classes = [] 
    for x in ws:
        if x[1] in db_dict: 
            for y in db_dict[x[1]]:
                if y in ok_entities:
                    classes.append(('db_' + y + '_' + x[0], x))
                    break 
    return classes

def load_dbpedia():
    # looks for 'dbpedia.json'; sky/sky/dbpedia.json
    with open(os.path.join(os.path.dirname(__file__), 'dbpedia.json')) as f:
        return json.load(f)



import langdetect

def get_language(tree, headers, domain = None):
    lang = None
    
    if headers and 'content-language' in headers:
        lang = headers['content-language']

    if lang is None and 'lang' in tree.attrib:
        lang = tree.attrib['lang']

    if lang is None:
        page_txt = ' '.join(tree.xpath('//p/text()')) + ' '.join(tree.xpath('//div/text()'))
        if not page_txt:
            page_txt = tree.text_content()
        lang = langdetect.detect(page_txt)

    # Note that this won't work with weird domain names. A mapping is needed from domain names to languages.
    if lang is None and domain is not None:
        lang = domain.split('.')[-1]    
        if lang == 'com':
            return 'en' 

    return lang[:2] or 'en'



"""A simple web crawler -- class implementing crawling logic."""
import traceback
import asyncio
import cgi
import os
from collections import namedtuple
import logging
import re
import time
import urllib.parse
import tldextract
import json
import shutil

from sky.scraper import Scraper
from sky.helper import makeTree

try:
    # Python 3.4.
    from asyncio import JoinableQueue as Queue
except ImportError:
    # Python 3.5.
    from asyncio import Queue

from asyncio import PriorityQueue


class JoinablePriorityQueue(Queue, PriorityQueue):
    pass

import aiohttp  # Install with "pip install aiohttp".

LOGGER = logging.getLogger(__name__)


def lenient_host(host):
    parts = host.split('.')[-2:]
    return ''.join(parts)


def is_redirect(response):
    return response.status in (300, 301, 302, 303, 307)


def slugify(value):
    url = re.sub(r'[^\w\s-]', '', re.sub(r'[-\s]+', '-', value)).strip().lower()
    return url[:-1] if url.endswith('/') else url


def extractDomain(url):
    tld = ".".join([x for x in tldextract.extract(url) if x])
    protocol = url.split('//', 1)[0]
    if 'file:' == protocol:
        protocol += '///'
    else:
        protocol += '//'
    return protocol + tld

FetchStatistic = namedtuple('FetchStatistic',
                            ['url',
                             'next_url',
                             'status',
                             'exception',
                             'size',
                             'content_type',
                             'encoding',
                             'num_urls',
                             'num_new_urls'])


class Crawler:
    """Crawl a set of URLs.

    This manages two sets of URLs: 'urls' and 'done'.  'urls' is a set of
    URLs seen, and 'done' is a list of FetchStatistics.
    """

    def __init__(self, config):
        self.loop = None
        self.seed_urls = None
        self.collections_path = None
        self.collection_name = None
        self.max_redirects_per_url = None
        self.max_saved_responses = 10000000
        self.max_tries_per_url = None
        self.max_workers = None
        self.crawl_required_regexps = []
        self.crawl_filter_regexps = []
        self.index_required_regexps = []
        self.index_filter_regexps = []
        self.login_data = {}
        self.login_url = None
        self.seen_urls = set()
        user_agent = config['user_agent'] if 'user_agent' in config else 'SkyBot v0.1'
        from_header = config['from'] if 'from' in config else "youremail@domain.com"
        self.headers = {'User-Agent': user_agent, 'From': from_header}
        for k, v in config.items():
            setattr(self, k, v)
        self.seen_urls = set(self.seen_urls)
        self.max_saved_responses = int(self.max_saved_responses)
        self.max_workers = min(int(self.max_workers), self.max_saved_responses)
        self.max_tries_per_url = int(self.max_tries_per_url)
        self.max_redirects_per_url = int(self.max_redirects_per_url)
        self.max_hops = int(self.max_hops)
        self.q = JoinablePriorityQueue(loop=self.loop)
        if 'queue' in config:
            for url in config['queue']:
                self.q.put_nowait((-1000, url, self.max_redirects_per_url))
        self.done = []
        self.root_domains = self.handle_root_of_seeds()
        self.t0 = time.time()
        self.t1 = None
        self.num_saved_responses = 0
        self.domain = extractDomain(self.seed_urls[0])
        self.file_storage_place = os.path.join(self.collections_path, self.collection_name)
        delete = False
        if delete and os.path.isdir(self.file_storage_place):
            shutil.rmtree(self.file_storage_place)
        if self.file_storage_place and not os.path.isdir(self.file_storage_place):
            os.makedirs(self.file_storage_place)

        self.session = aiohttp.ClientSession(headers=self.headers)

    @asyncio.coroutine
    def login(self):
        resp = yield from self.session.post(self.login_url, data=aiohttp.FormData(self.login_data))
        LOGGER.info('login in to url %r', self.login_url)
        print(resp.status)
        yield from resp.release()

    def handle_root_of_seeds(self):
        root_domains = set()
        for root in self.seed_urls:
            parts = urllib.parse.urlparse(root)
            host, _ = urllib.parse.splitport(parts.netloc)
            if host:
                if re.match(r'\A[\d\.]*\Z', host):
                    root_domains.add(host)
                else:
                    host = host.lower()
                    root_domains.add(lenient_host(host))
                self.add_url(0, root)
        if len(root_domains) > 1:
            raise Exception('Multiple Domains')
        return root_domains

    def close(self):
        """Close resources."""
        self.session.close()

    def host_okay(self, host):
        """Check if a host should be crawled.
        """
        host = host.lower()
        if host in self.root_domains:
            return True
        if re.match(r'\A[\d\.]*\Z', host):
            return False
        return self._host_okay_lenient(host)

    def _host_okay_lenient(self, host):
        """Check if a host should be crawled, lenient version.

        This compares the last two components of the host.
        """
        return lenient_host(host) in self.root_domains

    def record_statistic(self, fetch_statistic):
        """Record the FetchStatistic for completed / failed URL."""
        self.done.append(fetch_statistic)

    @asyncio.coroutine
    def save_response(self, text, response):
        with open(os.path.join(self.file_storage_place, slugify(response.url)), 'w') as f:
            json.dump({'url': response.url, 'html': text, 'headers': dict(response.headers)}, f)

    def should_crawl(self, url):
        if all([not re.search(x, url) for x in self.crawl_filter_regexps]):
            if (not self.crawl_required_regexps or
                    any([re.search(x, url) for x in self.crawl_required_regexps])):
                return True
        return False

    def should_save(self, url):
        if (not self.index_required_regexps or
                any([re.search(condition, url) for condition in self.index_required_regexps])):
            if all([not re.search(x, url) for x in self.index_filter_regexps]):
                return True
        return False

    @asyncio.coroutine
    def handle_response(self, response):
        """Return a FetchStatistic and list of links."""
        links = set()
        content_type = None
        encoding = None
        body = yield from response.read()

        if response.status == 200:
            num_allowed_urls = 0

            content_type = response.headers.get('content-type')
            pdict = {}

            if content_type:
                content_type, pdict = cgi.parse_header(content_type)

            encoding = pdict.get('charset', 'utf-8')

            if content_type in ('text/html', 'application/xml'):
                text = yield from response.text()

                current_url = response.url
                if self.should_save(current_url):
                    _ = yield from self.save_response(text, response)
                    self.num_saved_responses += 1
                    LOGGER.info('results: %r, CONVERTED url %r, ',
                                self.num_saved_responses, current_url)

                # Replace href with (?:href|src) to follow image links.
                urls = set(re.findall(r'''(?i)href=["']([^\s"'<>]+)''',
                                      text))

                for url in urls:
                    normalized = urllib.parse.urljoin(current_url, url)
                    defragmented, _ = urllib.parse.urldefrag(normalized)
                    if self.url_allowed(defragmented) and self.should_crawl(normalized):
                        if defragmented not in links and defragmented not in self.seen_urls:
                            num_allowed_urls += 1
                            links.add(defragmented)

                # visitable means: "urls that may be visit according to config"
                LOGGER.info('Queue: %r, FOUND ~%r visitable urls from %r, ',
                            self.q.qsize(), num_allowed_urls, current_url)

        stat = FetchStatistic(
            url=response.url,
            next_url=None,
            status=response.status,
            exception=None,
            size=len(body),
            content_type=content_type,
            encoding=encoding,
            num_urls=len(links),
            num_new_urls=len(links - self.seen_urls))

        return stat, links

    @asyncio.coroutine
    def fetch(self, prio, url, max_redirects_per_url):
        """Fetch one URL."""
        # Using max_workers since they are not being quit
        if self.num_saved_responses >= self.max_saved_responses:
            # NOT SURE IF THIS IS NEEDED
            return
        tries = 0
        exception = None
        while tries < self.max_tries_per_url:
            try:
                LOGGER.debug('GET url: ' + url)
                response = yield from asyncio.wait_for(
                    self.session.get(url, allow_redirects=False), 20)
                if tries > 1:
                    LOGGER.info('try %r for %r SUCCESS', tries, url)
                break
            except aiohttp.ClientError as client_error:
                LOGGER.info('try %r for %r RAISED %r', tries, url, client_error)
                exception = client_error
            except asyncio.TimeoutError as e:
                LOGGER.error('asyncio.TimeoutError for %r RAISED %r', url, e)
                exception = e
            except Exception as e:
                LOGGER.error('General error for %r RAISED %r', url, e)
                exception = e
            tries += 1
        else:
            # We never broke out of the loop: all tries failed.
            if self.max_tries_per_url > 1:
                LOGGER.error('%r FAILED after %r tries', url, self.max_tries_per_url)

            self.record_statistic(FetchStatistic(url=url,
                                                 next_url=None,
                                                 status=None,
                                                 exception=exception,
                                                 size=0,
                                                 content_type=None,
                                                 encoding=None,
                                                 num_urls=0,
                                                 num_new_urls=0))
            return

        if is_redirect(response):
            location = response.headers['location']
            next_url = urllib.parse.urljoin(url, location)
            self.record_statistic(FetchStatistic(url=url,
                                                 next_url=next_url,
                                                 status=response.status,
                                                 exception=None,
                                                 size=0,
                                                 content_type=None,
                                                 encoding=None,
                                                 num_urls=0,
                                                 num_new_urls=0))

            if next_url in self.seen_urls:
                yield from response.release()
                return
            if max_redirects_per_url > 0:
                LOGGER.info('REDIRECT to %r from %r', next_url, url)
                self.add_url(prio, next_url, max_redirects_per_url - 1)
            else:
                LOGGER.error('REDIRECT limit reached for %r from %r',
                             next_url, url)
        else:
            stat, links = yield from self.handle_response(response)
            self.record_statistic(stat)
            for link in links.difference(self.seen_urls):
                good = sum([x in link for x in self.index_required_regexps])
                bad = 10 * any([x in link for x in self.index_filter_regexps])
                prio = bad - good  # lower is better
                self.q.put_nowait((prio, link, self.max_redirects_per_url))

            self.seen_urls.update(links)
        yield from response.release()

    @asyncio.coroutine
    def work(self):
        """Process queue items forever."""
        while True:
            try:
                LOGGER.debug('get')
                prio, url, max_redirects_per_url = yield from asyncio.wait_for(self.q.get(), 30)
            except Exception as e:
                LOGGER.error('CRITICAL GET %r: stack %r', str(e), traceback.format_exc())
            try:
                LOGGER.debug('fetch')
                yield from self.fetch(prio, url, max_redirects_per_url)
            except Exception as e:
                LOGGER.error('CRITICAL FETCH %r: stack %r', str(e), traceback.format_exc())
            try:
                self.q.task_done()
            except Exception as e:
                LOGGER.error('CRITICAL DONE %r: stack %r', str(e), traceback.format_exc())

    def url_allowed(self, url):
        if url.endswith('.jpg') or url.endswith('.png'):
            return False
        parts = urllib.parse.urlparse(url)
        if parts.scheme not in ('http', 'https'):
            LOGGER.debug('SKIPPING non-http scheme in %r', url)
            return False
        host, _ = urllib.parse.splitport(parts.netloc)
        if not self.host_okay(host):
            LOGGER.debug('SKIPPING non-root host in %r', url)
            return False
        return True

    def add_url(self, prio, url, max_redirects_per_url=None):
        """Add a URL to the queue if not seen before."""
        if max_redirects_per_url is None:
            max_redirects_per_url = self.max_redirects_per_url
        LOGGER.debug('adding %r %r', url, max_redirects_per_url)
        self.seen_urls.add(url)
        self.q.put_nowait((prio, url, max_redirects_per_url))

    @asyncio.coroutine
    def crawl(self):
        """Run the crawler until all finished."""
        workers = [asyncio.Task(self.work(), loop=self.loop)
                   for _ in range(self.max_workers)]
        self.t0 = time.time()
        yield from self.q.join()
        print('seen urls {} done urls {}'.format(len(self.seen_urls), len(self.done)))
        self.t1 = time.time()
        for w in workers:
            w.cancel()

    def finish_leftovers(self):
        return False


class NewsCrawler(Crawler):

    def __init__(self, config):
        super(NewsCrawler, self).__init__(config)
        self.scraper = Scraper(config)
        self.template_complete = False
        self.data = {}
        self.templates_done = 0

    @asyncio.coroutine
    def save_response(self, text, response):
        # fucking mess
        try:
            # just let the indexer save the files as normal and also create a Template
            url = response.url
            tree = makeTree(text, self.scraper.domain)
            if self.templates_done < self.scraper.config['max_templates']:
                self.templates_done += 1
                self.scraper.domain_nodes_dict.add_template_elements(tree)
                self.scraper.url_to_headers_mapping[url] = dict(response.headers)
                self.data[url] = self.scraper.process(url, tree, False, ['cleaned'])
            else:
                # Let's try to do it in a tasked manner to remove existing ones and new ones
                # new one
                self.data[url] = self.scraper.process(url, tree, False, ['cleaned'])
        except Exception as e:
            try:
                fck = yield from response.text(encoding="cp1252")
                tree = makeTree(fck, self.scraper.domain)
                if self.templates_done < self.scraper.config['max_templates']:
                    self.templates_done += 1
                    self.scraper.domain_nodes_dict.add_template_elements(tree)
                    self.scraper.url_to_headers_mapping[url] = dict(response.headers)
                    self.data[url] = self.scraper.process(url, tree, False, ['cleaned'])
                else:
                    # Let's try to do it in a tasked manner to remove existing ones and new ones
                    # new one
                    self.data[url] = self.scraper.process(url, tree, False, ['cleaned'])
                LOGGER.warn("Solved UTF error with cp1252!")
            except Exception as e:
                LOGGER.error("RETRY CRITICAL ERROR IN SCRAPER for url %r: %r, stack %r",
                             url, str(e), traceback.format_exc())
        return

    def save_data(self, data):
        raise NotImplementedError('save_data has to be implemented')

    def save_bulk_data(self, data):
        raise NotImplementedError('save_bulk_data has to be implemented')

    def finish_leftovers(self):
        LOGGER.info('finish leftovers')
        if self.data:
            LOGGER.info('saving number of documents: ' + str(len(self.data)))
            LOGGER.info('saving status code: %r', self.save_bulk_data(self.data))
        return dict(self.scraper.domain_nodes_dict)



#!/usr/bin/env python3.4

"""A simple web crawler -- main driver program."""

import asyncio
import logging
import sys
import os

from sky.crawler.crawling import Crawler
from sky.crawler.reporting import report


def get_loop(select=False):
    if os.name == 'nt':
        from asyncio.windows_events import ProactorEventLoop
        loop = ProactorEventLoop()
    elif select:
        loop = asyncio.SelectorEventLoop()
    else:
        loop = asyncio.get_event_loop()
    return loop


def get_config(config, loop):
    for i in range(len(config['seed_urls'])):
        if '://' not in config['seed_urls'][i]:
            config['seed_urls'][i] = 'http://' + config['seed_urls'][i]

    config['loop'] = loop

    return config


def start(config, crawler_class=Crawler, save_data_result_fn=None, save_bulk_data_fn=None,
          logging_level=2):
    """Main program.

    Parse arguments, set up event loop, run crawler, print report.
    """

    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=logging_levels[min(logging_level, len(logging_levels) - 1)])

    loop = asyncio.SelectorEventLoop()

    asyncio.set_event_loop(loop)
    conf = get_config(config, loop)

    crawler = crawler_class(conf)

    if save_data_result_fn is not None:
        crawler.save_data = save_data_result_fn

    if save_bulk_data_fn is not None:
        crawler.save_bulk_data = save_bulk_data_fn

    if crawler.login_url:
        loop.run_until_complete(crawler.login())

    try:
        loop.run_until_complete(crawler.crawl())  # Crawler gonna crawl.
    except KeyboardInterrupt:
        sys.stderr.flush()
        print('\nInterrupted\n')
    except Exception as e:
        LOGGER.error('CRITICAL ERROR main loop exception: %r', e)
    finally:
        result = crawler.finish_leftovers()
        report(crawler)
        crawler.close()
        loop.close()
    return result



import re

locale_US = {'symbol' : r'\$',
             'currency' : r'dollar[s]*',
             'units' : [('million', 10**6),  ('m', 10**6), ('mn', 10**6), ('mil', 10**6),  
                        ('thousand', 10**3), ('k', 10**3), 
                        ('billion', 10**9), ('b', 10**9), ('bn', 10**9), 
                        ('cent', 0.01), 
                        ('\\b', 1), ('', 1)]}

class MoneyMatcher():
    def __init__(self, locale = None): 
        if locale is None:
            locale = locale_US
        self.unit_dict = {} 
        for k in locale['units']: 
            self.unit_dict[k[0]] = k[1]
            self.unit_dict[k[0].title()] = k[1]
            self.unit_dict[k[0].upper()] = k[1]
        units = '({})'.format('|'.join([x[0] for x in locale['units']])) + '\\b'
        number_regex = '([0-9]*[,.]*[0-9]+[,.]*)'
        optional_space = '[ ]*'
        self.symbol = re.compile(locale['symbol'] + optional_space + number_regex + optional_space + units, re.IGNORECASE)
        self.currency = re.compile(number_regex + optional_space + units + optional_space + locale['currency'], re.IGNORECASE) 
        
    def find(self, text, min_amount = 0, max_amount = 10**12):
        matches = []
        for m in self.symbol.finditer(text):
            matches.append([range(m.start(), m.end()), m.groups()])
            
        for m in self.currency.finditer(text):
            s  = m.start()
            e = m.end()
            for mm in matches:
                if s in mm[0]:
                    mm[0] = range(mm[0].start, e) 
                    break
                if e in mm[0]:
                    mm[0] = range(s, mm[0].stop)
                    break
            else:    
                matches.append([range(m.start(), m.end()), m.groups()])

        results = [(text[x[0].start:x[0].stop], self.convertMatchToValue(x[1]), x[0].start, x[0].stop) for x in matches]
                
        return [r for r in results if min_amount < r[1] < max_amount]
        
    def convertMatchToValue(self, match):
        value = match[0].replace(',', '.').strip('.')
        modifier = 1000 ** (len(re.findall(r'\.\d{3}[^0-9]', value)) + bool(re.search(r'\.\d{3}$', value))) 
        value = float(value) 
        unit_modifier = self.unit_dict[match[1]]
        return value * modifier * unit_modifier

def investment_annotation(title, body, money, entities, indicators = None, max_character_distance = 100): 
    # Create annotation of (company, money) if they are within max_characters_distance in title or body
    # Only if "invest", "fund" or "rais" are found nearby as well.
    total_content = title + ' ' + body
    if indicators is None:
        indicators = ['Invest', 'Fund', 'Rais']
    indicators = list(set(indicators + [x.lower() for x in indicators])) 
    entities_locations = [(e['text'], total_content.find(e['text'])) for e in entities if e['type'] in ['Person', 'Company']] 
    indicator_locations = [(x, total_content.find(x)) for x in indicators if total_content.find(x) > -1]
    me_combinations = {}
    for m in money:
        for e in entities_locations:
            for i in indicator_locations:
                if abs(e[1] - m[1]) < max_character_distance and abs(e[1] - i[1]) < max_character_distance:
                    if m not in me_combinations: 
                        me_combinations[m] = (e[0], abs(e[1] - m[1]))
                    elif abs(e[1] - m[1]) < me_combinations[m][1]: 
                        me_combinations[m] = (e[0], abs(e[1] - m[1]))

    return [{'company' : me_combinations[m][0], 'amount' : m[1]} for m in me_combinations]        



import os
import json
import translate

phrases = ['author', 'by', 'publi', 'write', 'written', 'info']

languages = ['aa','ab','af','ak','sq','am','ar','an','hy','as','av','ae','ay','az','ba','bm','eu','be','bn','bh','bi','bo','bs','br','bg','my','ca','cs','ch','ce','zh','cu','cv','kw','co','cr','cy','cs','da','de','dv','nl','dz','el','en','eo','et','eu','ee','fo','fa','fj','fi','fr','fr','fy','ff','ka','de','gd','ga','gl','gv','el','gn','gu','ht','ha','he','hz','hi','ho','hr','hu','hy','ig','is','io','ii','iu','ie','ia','id','ik','is','it','jv','ja','kl','kn','ks','ka','kr','kk','km','ki','rw','ky','kv','kg','ko','kj','ku','lo','la','lv','li','ln','lt','lb','lu','lg','mk','mh','ml','mi','mr','ms','mk','mg','mt','mn','mi','ms','my','na','nv','nr','nd','ng','ne','nl','nn','nb','no','ny','oc','oj','or','om','os','pa','fa','pi','pl','pt','ps','qt','qu','rm','ro','ro','rn','ru','sg','sa','si','sk','sk','sl','se','sm','sn','sd','so','st','es','sq','sc','sr','ss','su','sw','sv','ty','ta','tt','te','tg','tl','th','bo','ti','to','tn','ts','tk','tr','tw','ug','uk','ur','uz','ve','vi','vo','cy','wa','wo','xh','yi','yo','za','zh','zu']

answers = {}

for target in languages:
    answers[target] = {}
    for phrase in phrases: 
        try:
            answers[target][translate.translator('en', target, phrase)[0][0][0].lower()] = phrase
        except TypeError:
            print(target)    
            break

bad = []    
for a in answers:
    if answers[a]:
        for k in answers[a]:
            if k in phrases: 
                print('bad', a, k) 
                bad.append((a,k))

for a in bad:
    answers[a[0]].pop(a[1])
                
bad = []    
for a in answers:
    if not answers[a]:
        print('bad', a)
        bad.append(a)

for a in bad:
    answers.pop(a)        

fname = os.path.join(os.path.dirname(__file__), '/data/author_translation_table.json')    

with open(fname, 'w', encoding='utf8') as json_file:
    json.dump(answers, json_file, ensure_ascii=False)

    



import re
import json
import os
from pkg_resources import resource_filename

from sky.helper import get_text_and_tail

fname = os.path.abspath(resource_filename('sky.data', 'author_translation_table.json'))

with open(fname, encoding="utf-8") as f:
    author_translation_table = json.load(f)
    uppered = {x.title(): author_translation_table[x] for x in author_translation_table}
    author_translation_table.update(uppered)


def author_translation(txt, lang):
    if lang in author_translation_table:
        for month in author_translation_table[lang]:
            txt = txt.replace(month, author_translation_table[lang][month])
    return txt


def get_text_author(txt):
    res = re.findall(r"\b([A-Z][a-zA-Z' ]+)", txt)
    if res:
        for r in res:
            rsplit = r.split()
            if len(rsplit) < 6 and len(rsplit) > 1:
                return r.strip()
    return False


def get_author(tree, lang='en'):
    # Couple ways of matching:
    # - Both in meta and in text
    # - Node has one of the goods in it, text has it in it, and the case is authoric
    # - Node has one of the goods in it
    # - Text has one of the goods in it
    goods = ['author', 'by', 'publi', 'write', 'written', 'info']
    hard_authors = []
    meta_authors = []
    text_hard_authors = []
    text_soft_authors = []
    meta_nodes = tree.xpath('//head/meta')

    for option in goods:

        for meta in meta_nodes:
            if not any([option in a for a in meta.values()]):
                continue

            # dit gaat nog best helemaal fout! (ik moet atts nog chcken op goods)
            for attr in meta.attrib:
                author = get_text_author(author_translation(meta.attrib[attr], lang))
                if author:
                    meta_authors.append(author)

    for num, node in enumerate(tree.iter()):
        # hard author
        if not any([g in a for a in node.attrib.values() for g in goods]):
            for parent in node.iterancestors():
                attr_values = parent.attrib.values()
                if any([g in a for a in attr_values for g in goods]):
                    break
            else:
                continue
        tailtext = get_text_and_tail(node).strip()
        if tailtext and len(tailtext) < 200:
            if lang != 'en':
                tailtext = author_translation(tailtext, lang)
            hard_author = get_text_author(tailtext)
            if hard_author:
                hard_authors.append((num, hard_author))

    for num, node in enumerate(tree.iter()):
        tailtext = get_text_and_tail(node).strip()
        if tailtext and len(tailtext) < 200:
            res = re.findall(r"(author|Author|AUTHOR)[:;]* ([A-Z][a-zA-Z' ]+)", tailtext)
            if res:
                res = res[0]
                if res in meta_authors:
                    text_hard_authors.append((res, num))
                else:
                    text_soft_authors.append((res, num))
            else:
                res = re.findall(r"\b(by|By|BY)[:;]* ([A-Z][a-zA-Z' ]+)", tailtext)
                if res:
                    res = res[0]
                    if res in meta_authors:
                        text_hard_authors.append((res, num))
                    else:
                        text_soft_authors.append((res, num))

    hardest_authors = []
    not_hardest_authors = []
    for num, ha in hard_authors:
        if ha in meta_authors:
            hardest_authors.append((ha, num))
        else:
            not_hardest_authors.append((ha, num))

    meta_authors = set(meta_authors)

    return hardest_authors, not_hardest_authors, text_hard_authors, text_soft_authors, meta_authors
